/*******************************************************************\

Module: Bounded Model Checking for ANSI-C + HDL

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#ifndef CPROVER_CBMC_BMC_H
#define CPROVER_CBMC_BMC_H

#include <list>
#include <map>

#include <util/hash_cont.h>
#include <util/options.h>

/*** added by sbjoshi ***/
#include <glue/mhs.h>
#include <glue/wmm_repair.h>
#include <glue/glue_transformation.h>
#include <glue/glue_util.h>
#include <util/time_stopping.h>

#include <solvers/prop/prop.h>
#include <solvers/prop/prop_conv.h>
#include <solvers/sat/cnf.h>
#include <solvers/sat/satcheck.h>
#include <solvers/smt1/smt1_dec.h>
#include <solvers/smt2/smt2_dec.h>
#include <langapi/language_ui.h>
#include <goto-symex/symex_target_equation.h>

#include "symex_bmc.h"

typedef enum {SATOMIC=1,WATOMIC=2,TSOSC=3,PSOTSO=4,PSOSC=5} repairmodet;


class bmct:public messaget
{
public:
  bmct(
    const optionst &_options,
    const symbol_tablet &_symbol_table,
    message_handlert &_message_handler):
    messaget(_message_handler),
    options(_options),
    ns(_symbol_table, new_symbol_table),
    equation(ns),
    symex(ns, new_symbol_table, equation),
    ui(ui_message_handlert::PLAIN)
  {
    symex.constant_propagation=options.get_bool_option("propagation");
  }
 
  virtual bool run(const goto_functionst &goto_functions);
  virtual ~bmct() { }

  // additional stuff   
  expr_listt bmc_constraints;  
  
  friend class cbmc_satt;
  friend class hw_cbmc_satt;
  friend class counterexample_beautification_greedyt;
  
  void set_ui(language_uit::uit _ui) { ui=_ui; }
  /** added by sbjoshi **/
   virtual bool sa_repair();
   virtual bool wa_repair();
   virtual bool wmm_repair();
   //member required for --repair option

   satomic_transformationt satomic_repair;
   watomic_transformationt watomic_repair;
  // typedef std::list<literalt> literalst;
    typedef mhst<unsigned int> mhsit;
   typedef std::vector<literalst> guard_to_literal_mapt;
   guard_to_literal_mapt guard_to_literal_map;

   typedef std::vector<literalst> wbguard_to_literal_mapt;
   wbguard_to_literal_mapt wbguard_to_literal_map;
  
   //sbjoshi
   void reset_bmc()
   {
	   new_symbol_table.clear();
	   equation.clear();

   }
   //remove after debugging -- added by sbjoshi
   void print_cs_evaluation(const prop_conv_solvert& prop_conv,const std::set<unsigned int>& s);

   void report_reordering_fix(const std::set<unsigned int>& groups,literal2symbolt& literal2symbol,
		   orderingliteralst& oliterals);
   fine_timet formula_build_time_start; //added by sbjoshi
protected:

  const optionst &options;  
  symbol_tablet new_symbol_table;
  namespacet ns;
  symex_target_equationt equation;
  symex_bmct symex;
 
  // use gui format
  language_uit::uit ui;
  
  virtual decision_proceduret::resultt
    run_decision_procedure(prop_convt &prop_conv);
    
  virtual bool decide(prop_convt &prop_conv);
    
  // the solvers we have
  virtual bool decide_default();
  virtual bool decide_bv_refinement();
  virtual bool decide_aig();
  virtual bool decide_smt1();
  virtual bool decide_smt2();
  smt1_dect::solvert get_smt1_solver_type() const;
  smt2_dect::solvert get_smt2_solver_type() const;
  virtual void smt1_convert(smt1_dect::solvert solver, std::ostream &out);
  virtual void smt2_convert(smt2_dect::solvert solver, std::ostream &out);
  virtual bool write_dimacs();
  virtual bool write_dimacs(std::ostream &out);
  
  // unwinding
  virtual void setup_unwind();

  virtual void do_unwind_module(
    decision_proceduret &decision_procedure);
  void do_conversion(prop_convt &solver);
  
  prop_convt *solver_factory();

  virtual void show_vcc();
  virtual bool all_claims(
    const goto_functionst &goto_functions,
    prop_convt &solver);
  virtual void show_vcc(std::ostream &out);
  virtual void show_program();
  virtual void report_success();
  virtual void report_failure();

  virtual void error_trace(
    const prop_convt &prop_conv);
  
  // vacuity checks
  void cover_assertions(
    const goto_functionst &goto_functions,
    prop_convt &solver);

  /*** added by sbjoshi **/

  symbol2ordert symbol2order;
  /*typedef std::map<irep_idt,unsigned int> symbol2grpt;
 // symbol2grpt symbol2grp;
  typedef std::vector<literalst> orderingliteralst;
 // orderingliteralst oliterals;
  typedef std::map<literalt,irep_idt> literal2symbolt;
 // literal2symbolt literal2symbol;*/

  void group_from_ordering(symbol2grpt& symbol2grp,sym2grpsymt& sym2grpsym);
  //void populate_oliterals_map(prop_conv_solvert& prop_conv);
  void populate_oliterals_map(prop_conv_solvert& prop_conv, literal2symbolt& literal2symbol,
  		symbol2grpt& symbol2grp, orderingliteralst& oliterals,grp2grplitt& grp2grplit);
    void populate_guard_literal_map(const prop_conv_solvert& prop_conv,atomicityt a);
   // void populate_wbguard_literal_map(prop_convt& prop_conv,atomicityt a);
  void initialize_guard_literal_map(unsigned int size,atomicityt a);
 // void initialize_wbguard_literal_map(unsigned int size);
  void print_guard_to_literal_map(std::ostream &out,atomicityt a);
 // void print_wbguard_to_literal_map(std::ostream &out);
 
  void update_atomicity_constraints(symex_target_equationt &target,
  prop_conv_solvert &prop_conv,
  const namespacet &ns,
  goto_tracet &goto_trace,mhsit& m,atomicityt a);
  void add_atomicity_assumptions(prop_conv_solvert& prop_conv, std::set<unsigned int>& ban_cs,atomicityt a) const;
  //void add_reordering_assumptions(prop_conv_solvert& prop_conv,std::set<unsigned int>& ban_reord) const;
  void add_reordering_assumptions(prop_conv_solvert& prop_conv,
      std::set<unsigned int>& ban_reord,grp2grplitt& grp2grplit, bvt& prev_assumptions) const;
  void add_reordering_memorax_constraint(propt& prop,
  		 std::set<unsigned int>& gtaken, grp2grplitt& grp2grplit) const;
  //void update_ordering_constraints(prop_conv_solvert &prop_conv, mhsit& m);
  void update_ordering_constraints(
     prop_conv_solvert &prop_conv, mhsit& m,literal2symbolt& literal2symbol,
     symbol2grpt& symbol2grp,grp2grplitt& grp2grplit, bvt& prev_assumptions);
  void update_ordering_constraints(
     prop_conv_solvert &prop_conv, propt& prop, mhsit& m,literal2symbolt& literal2symbol,
     symbol2grpt& symbol2grp,grp2grplitt& grp2grplit);
 // void get_reordering_from_prop(prop_conv_solvert& prop_conv,std::set<unsigned int>& s);
  void get_reordering_from_prop(prop_conv_solvert& prop_conv,
       std::set<unsigned int>& s, const literal2symbolt& literal2symbol,
       symbol2grpt& symbol2grp);

};

#endif
