
#ifndef CPROVER_GLUE_MHS_H
#define CPROVER_GLUE_MHS_H
#include <vector>
#include <map>
#include <set>
#include <limits>
#include <algorithm>
#include <iterator>
#include <solvers/prop/prop.h>
#include <cassert>

#include <solvers/prop/cardinality.h>
#include <util/threeval.h>

#ifdef DEBUG
#include <iostream>
#endif
typedef std::set<unsigned int> uisett;
/*****************************************************************************\
 * Class : mhst
 *
 * Purpose : A generic type to find the minimum hitting set over a collection
 * C of sets of objects of type T
 *
 * A hitting set is a set H such that \forall s\in C H \cap s \neq {}, in
 * other words, a hitting set has a non-empty intersection with all sets
 * in C. The smallest such hitting set is called the minimum hitting set
 *
 * Note : There can be multiple minimum hitting set for a collection
 *
 * Methods used : It uses a greedy algorithm to compute the minimum hitting
 * set and then get_mhs_from_solver() uses a constraint solver to verify
 * or further reduce the size of the hitting set until optimum solution is 
 * found
 *
 * The greedy algorithm is as follows :
 * Pick an element e which hits (is contained in) maximum number of 
 * uncovered (sets from which no element has been picked yet) sets.
 * Repeat until all sets are covered. This algorithm has an approximation
 * factor of log n. However, for nicely behaved structures like programs,
 * the collection generated is very well-behaved and greedy algorithm provides
 * optimum solution in the most cases
 *
 *
 * ***************************************************************************/
template <typename T>
class mhst
{
  public :
    typedef typename std::set<T> sett;
    typedef typename std::vector<sett > collectiont;
    typedef typename std::map<T,unsigned int> freq_mapt;
    typedef typename std::map<T,std::set<unsigned int> > elem_to_sett;
    typedef typename std::map<T,literalt> elem_to_litt;
    typedef typename std::map<literalt,T> lit_to_elemt;

  protected :
    // map from elements to the sets it hits
    elem_to_sett elem_to_set;
    //collection C
    collectiont collection;

    // map from elements to the number of set it hits
    freq_mapt freq_map;

    // solution contains the minimum hitting set - used for
    // memoization
    sett solution;

    // flag that indicates if solution contains a valid hitting set
    // for C
    bool contains_valid_mhs;

    // copy one set to another 
    void make_set_copy(const sett& from,sett& to)
    {
      for(typename sett::iterator it=from.begin();it!=from.end();it++)
      {
        to.insert(*it);
      }
    }
  public:
    // initially the solution does not contain a valid mhs
    mhst() {contains_valid_mhs=false;}

    // add a set s into the collection C
    void add_to_collection(sett& s)
    {
      //since we are changing C, the solution may not hold a 
      //valid mhs
        
      contains_valid_mhs=false;
      collection.push_back(s);
      typename std::vector<sett>::size_type vsize = collection.size();

      //update the freq_map, elem_to_set
      for(typename sett::const_iterator it=s.begin(); it!=s.end();it++)
      {

        if(freq_map.find(*it)!=freq_map.end())
        {
          freq_map[*it]++;
        }
        else
        {
          freq_map[*it] = 1;
        }

        if(elem_to_set.find(*it)!=elem_to_set.end())
        {
          elem_to_set[*it].insert(vsize-1);
        }
        else
        {
          std::set<unsigned int> s;
          s.insert(vsize-1);
          elem_to_set[*it]=s;
        }
      }


    }
#ifdef DEBUG


    void print_freq_map(std::ostream& os,const freq_mapt& fmap) const
    {

      os << "Freq map begin " << std::endl;
      for(typename freq_mapt::const_iterator it=fmap.begin();
          it!=fmap.end();it++)
      {
        os << it->first << " : " << it->second << " :: ";
      }
      os << std::endl << "freq map end" << std::endl;
    }

    void print_collection(std::ostream& os) const
    {
      os << "Printing collection start" << std::endl;

      for(typename collectiont::const_iterator it=collection.begin();
          it!=collection.end();it++)
      {
        for(typename sett::const_iterator sit=it->begin();sit!=it->end();sit++)
        {
          os << (*sit) << " ";
        }
        os << std::endl;

      }
      os << "printing collection end" << std::endl;
    }
#endif

    // element e has been picked to be in solution, so update fmap
    // to reflect the frequencies for the uncovered sets
    virtual void update_freq_map(freq_mapt& fmap, const T&e,
    		const std::set<unsigned int>& covered_sets)
    {

      const std::set<unsigned int>& covered_by_elem = elem_to_set[e];

      std::set<unsigned int> sets_to_be_covered;

      uisett::const_iterator f1=covered_by_elem.begin(),l1=covered_by_elem.end();
      uisett::const_iterator f2=covered_sets.begin(),l2=covered_sets.end();
      while(f1!=l1 && f2!=l2)
      {
    	  if((*f1) < (*f2)) {sets_to_be_covered.insert(*f1);f1++;}
    	  else if((*f2) < (*f1)) {f2++;}
    	  else {f1++;}
      }
      while(f1!=l1)
      {
    	  sets_to_be_covered.insert(*f1);f1++;
      }

      for(std::set<unsigned int>::const_iterator it = sets_to_be_covered.begin();
          it!=sets_to_be_covered.end();it++)
      {
        const sett& curr_set = collection[*it];

        for(typename sett::const_iterator cit=curr_set.begin();
            cit!=curr_set.end();cit++)
        {
          if(fmap[*cit]>1)
          {
            fmap[*cit]--;
          }
          else
          {
            fmap[*cit]=0;
          }

        }
      }
    }

    // pick an element e which is hitting the most number of uncovered sets
    virtual void get_max_efficient_elem(const freq_mapt& fmap,T& e) const
    {
      unsigned int max_val = std::numeric_limits<unsigned int>::min();

      for(typename freq_mapt::const_iterator it = fmap.begin();
          it!=fmap.end();it++)
      {
        if(it->second > max_val)
        {
          e = it->first;
          max_val = it->second;
        }
      }

    }

    // returns true if sol is a hitting set even if e is removed from sol
    virtual bool is_cover(const sett& sol,const  T& e) 
    {
      std::set<unsigned int> covered;
      for(typename sett::const_iterator it=sol.begin();it!=sol.end();it++)
      {
        if(*it == e)
        { continue;
        }  

        for(std::set<unsigned int>::const_iterator cit=elem_to_set[*it].begin();
            cit!=elem_to_set[*it].end();cit++)
        {
          covered.insert(*cit);
        }
      }
      if(covered.size()==collection.size())
      {
        return true;
      }
      else
      {
        return false;
      }
    }

    //create a copy of member freq_map into fmap
    void get_freq_map(freq_mapt& fmap) const
    {
      for(typename freq_mapt::const_iterator it=freq_map.begin();
          it!=freq_map.end();it++)
      {
        fmap[it->first] = it->second;
      }
    }

    // returns true if sol is a hitting set for C
    bool is_cover(const sett& sol, std::set<unsigned int>& covered)
    {
      //std::set<unsigned int> covered;
      for(typename sett::const_iterator it=sol.begin();it!=sol.end();it++)
      {


        for(std::set<unsigned int>::const_iterator cit=elem_to_set[*it].begin();
            cit!=elem_to_set[*it].end();cit++)
        {
         covered.insert(*cit);

#ifdef DEBUG
         std::cout << (*cit) << " is covered" << std::endl;
#endif

        }
      }

      if(covered.size()==collection.size())
      {
        return true;
      }
      else
      {
        return false;
      }
    }


    // try removing each element of sol and see if it is still
    // a hitting set of C, if yes, remove the element from sol
    void make_minimal(sett& sol) 
    {
      sett elem_to_be_removed;
      for(typename sett::const_iterator it=sol.begin();it!=sol.end();it++)
      {
        if(is_cover(sol,*it))
        {
          sol.erase(*it);
        }


      }


    }


    // computes mhs of C using the greedy algorithm and then
    // makes it minimal
    virtual void get_mhs(sett& mhs)
    {
      mhs.clear();
      if(contains_valid_mhs)
      {
        make_set_copy(solution,mhs);
        return;

      }
      /*
#ifdef DEBUG
      print_collection(std::cout);
#endif
*/
      std::set<unsigned int> covered_sets;
      freq_mapt new_fmap;
      get_freq_map(new_fmap);
/*
      #ifdef DEBUG
      std::cout << "Freq map ";
      print_freq_map(std::cout,new_fmap);

#endif*/
      while(!is_cover(mhs,covered_sets))
      {
        T e;
        get_max_efficient_elem(new_fmap,e);
        /*
#ifdef DEBUG
        std::cout << "element chosen " << e << std::endl;
#endif*/
        mhs.insert(e);
        update_freq_map(new_fmap,e,covered_sets);
        /*
#ifdef DEBUG
        std::cout <<"Update fmap now ";
        print_freq_map(std::cout,new_fmap);
#endif*/
        covered_sets.clear();

      }
      make_minimal(mhs);

      solution.clear();
      make_set_copy(mhs,solution);
      contains_valid_mhs=true;
    }

    // add C as a formula in the solver (prop)
    // add a literal for every element and maintain the 
    // mapping in elem_to_lit and lit_to_elem
    void add_collection_to_solver(propt& prop,bvt& elements,
        elem_to_litt& elem_to_lit,lit_to_elemt& lit_to_elem)
    {
      lit_to_elem.clear();
      elem_to_lit.clear();
      for(typename collectiont::iterator cit=collection.begin();
          cit!=collection.end();cit++)
      {
        bvt clause;
        for(typename sett::iterator sit=cit->begin();
            sit!=cit->end();sit++)
        {
          literalt l;
          typename elem_to_litt::iterator eit=elem_to_lit.find(*sit);
          if(eit!=elem_to_lit.end())
          {
            l=eit->second;
          }  
          else
          {
            l=prop.new_variable();
            if(l.sign()){l.invert();}
            elem_to_lit[*sit]=l;
            lit_to_elem[l]=*sit;
            elements.push_back(l);

          }
          clause.push_back(l);

        }
        prop.lcnf(clause);
      }

    }

    // from the solver, get the elements picked as the solution
    void get_positive_elements(const propt& prop, sett& elements,
        lit_to_elemt& lit_to_elem)
    {

      for(typename lit_to_elemt::const_iterator lit =lit_to_elem.begin();
          lit!=lit_to_elem.end();lit++)
      {
        if(prop.l_get(lit->first).is_true())
        {
          elements.insert(lit->second);
        }
      }
    }

    //use the solver (prop) to get an mhs
    virtual bool get_mhs_from_solver(sett& mhs,propt& prop)
    {
      //first get a candidate from greedy algorithm
      get_mhs(mhs);    

      //if solver does not support incremental solving
      //bail out
      if(!prop.has_set_assumptions()) return false;

      unsigned int k = mhs.size();
#ifdef DEBUG
      std::cout << "Approx mhs is " << k << std::endl;
#endif
      //if the greedy algorithm has provided mhs of size 1
      //that is as good as it can get
      if(k<=1) return true;  


      // since we already have a solution, see if a smaller
      // solution exist
      k--;
      //start from k and keep decreasing k until solver
      //returns unsatisfiable--telling that no hitting
      //set of size k exist
      //
      //
      //One can use binary search as well, however, that
      //would involve posing more unsatisfiable queries,
      //In this way, there would be only single unsatisfiable
      //query at the end
      bool done=false;

      bvt elements;
      elem_to_litt e2l;
      lit_to_elemt l2e;
      add_collection_to_solver(prop,elements,e2l,l2e);
      do
      {

        //encode the constraint to ask the solver if 
        //there exist a hitting set of size at most k 
        sequential_encodingt ce(prop);  
        formulat formula;

        ce.atmostk(elements,k,formula);
        bvt tmplits;
        tmplits.push_back(ce.get_lit_for_formula(formula));
        prop.set_assumptions(tmplits);
        propt::resultt result=prop.prop_solve();
        switch(result)
        {
          case propt::P_SATISFIABLE: mhs.clear();
                                     get_positive_elements(prop,mhs,l2e);
                                     // solver may return a set
                                     // smaller than k (if we are lucky)
                                     k=mhs.size()-1;
#ifdef DEBUG
      std::cout << "Solver mhs is " << k+1 << std::endl;
#endif

                                     break;
                                     //bailout if unsatisfiable
          case propt::P_UNSATISFIABLE: done=true;break;
          default: return false;
        }


      }while(k!=1 && !done);

      return true;
    }
};
/*****************************************************************************\
 * class : wmhst
 *
 * Purpose : A generic class to find a minimum cost hitting set
 * for a collection C of sets of objects of type T. minimum cost hitting
 * set is a hitting set where the cost is the smallest. In this version
 * each element has a cost associated with it
 * **************************************************************************/
template <typename T>
class wmhst : public mhst<T>
{  typedef typename std::set<T> sett;
  typedef typename std::vector<sett > collectiont;
  typedef typename std::map<T,unsigned int> freq_mapt;
  typedef typename std::map<T,std::set<unsigned int> > elem_to_sett;
  typedef typename std::map<T,literalt> elem_to_litt;
  typedef typename std::map<literalt,T> lit_to_elemt;

  typedef typename std::map<T,unsigned int> elem_to_wtt;
  //map from elements to weight
  elem_to_wtt elem_to_wt;

  public : 
  //by default all the elements have weight 1, so add
  //a weight if more than 1
  bool add_weight_to_elem(T& elem, unsigned int weight)
  {
    if(weight<=1 || 
        mhst<T>::freq_map.find(elem)==mhst<T>::freq_map.end())
    { return false;  }

    elem_to_wt[elem]=weight;
    //weight has changed so the solution might not
    //contain a valid mhs
    mhst<T>::contains_valid_mhs=false;
    return true;
  }

  // pick an element e such that the ratio of the 
  // number of sets it hits to its weight is the highest
  void get_max_efficient_elem(const freq_mapt& fmap,T& e) const
  {
    float max_val = std::numeric_limits<float>::min();

    for(typename freq_mapt::const_iterator it = fmap.begin();
        it!=fmap.end();it++)
    {
      float hit_ratio;
      typename elem_to_wtt::const_iterator wit=elem_to_wt.find(it->first);
      if(wit!=elem_to_wt.end())
      {
        //multiplying by 100 to avoid getting too small numbers 
        //to avoid underflow!! just in case
        hit_ratio = (100*((float)it->second))/(float)wit->second;

      }
      else
      {
        hit_ratio = (float)it->second;
      }
      if(hit_ratio > max_val)
      {
        e = it->first;
        max_val = hit_ratio;
      }
    }

  }

  // returns the sum of weights of all the elements in elems
  unsigned int get_weight(const sett& elems)
  {
    unsigned int sum=0;
    for(typename sett::const_iterator sit=elems.begin();
        sit!=elems.end();sit++)
    {
      typename elem_to_wtt::const_iterator wit=elem_to_wt.find(*sit);

      if(wit!=elem_to_wt.end())
      {
        sum+=wit->second;
      }
      else
      {
        sum++;
      }
    }

    return sum;
  }
// invoke solver to find the minimum cost hitting set
  bool get_mhs_from_solver(sett& mhs,propt& prop)
  { 
    //weighted version does not scale for solver
    // binomial encoding and subset construction goes too far
    // for higher weights

    //get a candidate from approximation algorithm
    this->get_mhs(mhs);    

    //bail out if the solver does not support incremental
    //solving
    if(!prop.has_set_assumptions()) return false;

    unsigned int k = get_weight(mhs);
    if(k<=1) return true;  

    //do a binary search to find the optimum k for which 
    //a hitting set exist. Binary search makes sense here
    //because weights could be large so the solution from
    //approximation algorithm may deviate too much and decrementing
    //k one by one may require too many queires
    unsigned int max_weight=k;
    unsigned int min_weight=1;
    bvt elements;
    elem_to_litt e2l;
    lit_to_elemt l2e;
    add_collection_to_solver(prop,elements,e2l,l2e);
    do
    {
      k = (max_weight+min_weight) >> 1;
      sequential_encodingt ce(prop);  
      formulat formula;
      ce.atmostk(elements,k,formula);
      literalt l=ce.get_lit_for_formula(formula);
      bvt assumptions;
      assumptions.push_back(l);
      prop.set_assumptions(assumptions);
      propt::resultt result=prop.prop_solve();
      switch(result)
      {
        case propt::P_SATISFIABLE: mhs.clear();
                                   mhst<T>::get_positive_elements(prop,mhs,l2e);
                                   max_weight=get_weight(mhs);
                                   break;
        case propt::P_UNSATISFIABLE: 
                                   min_weight=k;break;
        default: return false;
      }


// bail out when diff between min_weight and max_weight
// is less than 1, since min_weight is already UNSAT
// we have found the optimum answer!
    } while((max_weight-min_weight)>1);

    return true;
  }

  //for every element e have a literal l and remember the mapping
  //Also for elements e, make wight(e)-1 copies of auxiliary vars
  //and whenever l is true, all of these aux vars has to be true
  //This is saying that when you pick an element, you have to consider
  //the entire weight of the elem. This technique may not scale well
  //for very large weights
  void add_collection_to_solver(propt& prop,bvt& literals,
      elem_to_litt& elem_to_lit,
      lit_to_elemt& lit_to_elem)
  {
    lit_to_elem.clear();
    elem_to_lit.clear();
    for(typename collectiont::iterator cit=mhst<T>::collection.begin();
        cit!=mhst<T>::collection.end();cit++)
    {
      bvt clause;
      for(typename sett::iterator sit=cit->begin();
          sit!=cit->end();sit++)
      {
        literalt l;
        typename elem_to_litt::iterator eit=elem_to_lit.find(*sit);
        if(eit!=elem_to_lit.end())
        {
          l=eit->second;
        }  
        else
        {
          l=prop.new_variable();
          if(l.sign()){l.invert();}
          elem_to_lit[*sit]=l;
          lit_to_elem[l]=*sit;
          literals.push_back(l);

        }
        clause.push_back(l);

      }
      prop.lcnf(clause);
    }

    for(typename elem_to_wtt::const_iterator wit=elem_to_wt.begin();
        wit!=elem_to_wt.end();wit++)
    {
      bvt grouplits;
      typename elem_to_litt::iterator eit=elem_to_lit.find(wit->first);
      assert(eit!=elem_to_lit.end() && 
          "Literal for the element should have been added already!");

      // if weight is w_i for e_i, add w_i-1 auxiliary variables
      // then add l_i -> \and w_i-1 (if the literal corresponding to the 
      // element is true, so are all the corresponding auxiliary variables)
      // These auxiliary variables will be passed for atmostk constraints 
      // for weight
      for(unsigned int i=1;i<wit->second;i++)
      {
        grouplits.push_back(prop.new_variable());

      }  

      literalt gl = prop.land(grouplits);
      literalt l = prop.limplies(eit->second,gl);
      bvt tmp_clause;
      tmp_clause.push_back(l);
      prop.lcnf(tmp_clause);

      literals.insert(literals.begin(),grouplits.begin(),grouplits.end());


    }

  }



};


#endif


